@@ -2,7 +2,7 @@ use std::any::TypeId;
 use std::cell::{Ref, RefMut};
 use std::marker::PhantomData;
 use std::mem::{transmute, ManuallyDrop};
-use std::ptr::{null, null_mut};
+use std::ptr::NonNull;
 use std::slice::Iter;
 
 use crate::{
@@ -94,7 +94,7 @@ where
     tag_gen: (u32, u32),
     types: Vec<(usize, <S::Fetch as Fetch<'static>>::Ty)>,
     refs: Vec<ManuallyDrop<<S::Fetch as Fetch<'static>>::Ref>>,
-    ptrs: Vec<(u32, <S::Fetch as Fetch<'static>>::Ptr)>,
+    ptrs: Vec<Option<<S::Fetch as Fetch<'static>>::Ptr>>,
 }
 
 unsafe impl<S> Send for Query<S> where S: QuerySpec {}
@@ -155,7 +155,7 @@ where
 
         let types: &'w [(usize, <S::Fetch as Fetch<'w>>::Ty)] = unsafe { transmute(&*self.types) };
         let refs: &'w mut Vec<<S::Fetch as Fetch<'w>>::Ref> = unsafe { transmute(&mut self.refs) };
-        let ptrs: &'w mut Vec<(u32, <S::Fetch as Fetch<'w>>::Ptr)> =
+        let ptrs: &'w mut Vec<Option<<S::Fetch as Fetch<'w>>::Ptr>> =
             unsafe { transmute(&mut self.ptrs) };
 
         for (idx, ty) in types {
@@ -232,7 +232,7 @@ where
     world: &'w World,
     types: &'w [(usize, <S::Fetch as Fetch<'w>>::Ty)],
     refs: &'w mut Vec<<S::Fetch as Fetch<'w>>::Ref>,
-    ptrs: &'w mut Vec<(u32, <S::Fetch as Fetch<'w>>::Ptr)>,
+    ptrs: &'w mut Vec<Option<<S::Fetch as Fetch<'w>>::Ptr>>,
 }
 
 impl<S> QueryRef<'_, S>
@@ -248,7 +248,7 @@ where
             archetypes: &self.world.archetypes,
             idx: 0,
             len: 0,
-            ptr: S::Fetch::null(),
+            ptr: S::Fetch::dangling(),
         }
     }
 
@@ -276,20 +276,19 @@ where
     /// ```
     pub fn map<'q>(&'q mut self) -> QueryMap<'q, S> {
         let types: &'q [(usize, <S::Fetch as Fetch<'q>>::Ty)] = unsafe { transmute(self.types) };
-        let ptrs: &'q mut Vec<(u32, <S::Fetch as Fetch<'q>>::Ptr)> =
+        let ptrs: &'q mut Vec<Option<<S::Fetch as Fetch<'q>>::Ptr>> =
             unsafe { transmute(&mut *self.ptrs) };
 
         ptrs.clear();
 
-        ptrs.resize(self.world.archetypes.len(), (0, S::Fetch::null()));
+        ptrs.resize(self.world.archetypes.len(), None);
 
         for (idx, ty) in types {
             let archetype = &self.world.archetypes[*idx];
 
-            let len = archetype.len();
             let ptr = unsafe { S::Fetch::base_pointer(archetype, *ty) };
 
-            ptrs[*idx] = (len, ptr);
+            ptrs[*idx] = Some(ptr);
         }
 
         QueryMap {
@@ -370,7 +369,7 @@ where
     S: QuerySpec,
 {
     entities: &'q [EntityMetadata],
-    ptrs: &'q [(u32, <S::Fetch as Fetch<'q>>::Ptr)],
+    ptrs: &'q [Option<<S::Fetch as Fetch<'q>>::Ptr>],
 }
 
 impl<'q, S> QueryMap<'q, S>
@@ -382,16 +381,10 @@ where
         let meta = self.entities[ent.id as usize];
         assert_eq!(ent.gen, meta.gen, "Entity is stale");
 
-        let (len, ptr): &'m (u32, <S::Fetch as Fetch<'m>>::Ptr) =
+        let ptr: &'m Option<<S::Fetch as Fetch<'m>>::Ptr> =
             unsafe { transmute(&self.ptrs[meta.ty as usize]) };
 
-        if meta.idx < *len {
-            let val = unsafe { S::Fetch::deref(*ptr, meta.idx) };
-
-            Some(val)
-        } else {
-            None
-        }
+        ptr.map(|ptr| unsafe { S::Fetch::deref(ptr, meta.idx) })
     }
 }
 
@@ -412,7 +405,7 @@ pub unsafe trait Fetch<'q> {
     unsafe fn borrow(archetype: &'q Archetype, ty: Self::Ty) -> Self::Ref;
     unsafe fn base_pointer(archetype: &'q Archetype, ty: Self::Ty) -> Self::Ptr;
 
-    fn null() -> Self::Ptr;
+    fn dangling() -> Self::Ptr;
     unsafe fn deref(ptr: Self::Ptr, idx: u32) -> Self::Item;
 }
 
@@ -431,7 +424,7 @@ where
 {
     type Ty = usize;
     type Ref = Ref<'q, ()>;
-    type Ptr = *const C;
+    type Ptr = NonNull<C>;
 
     type Item = &'q C;
 
@@ -444,15 +437,15 @@ where
     }
 
     unsafe fn base_pointer(archetype: &'q Archetype, ty: Self::Ty) -> Self::Ptr {
-        archetype.base_pointer::<C>(ty)
+        NonNull::new_unchecked(archetype.base_pointer::<C>(ty))
     }
 
-    fn null() -> Self::Ptr {
-        null()
+    fn dangling() -> Self::Ptr {
+        NonNull::dangling()
     }
 
     unsafe fn deref(ptr: Self::Ptr, idx: u32) -> Self::Item {
-        &*ptr.add(idx as usize)
+        &*ptr.as_ptr().add(idx as usize)
     }
 }
 
@@ -471,7 +464,7 @@ where
 {
     type Ty = usize;
     type Ref = RefMut<'q, ()>;
-    type Ptr = *mut C;
+    type Ptr = NonNull<C>;
 
     type Item = &'q mut C;
 
@@ -488,15 +481,15 @@ where
     }
 
     unsafe fn base_pointer(archetype: &'q Archetype, ty: Self::Ty) -> Self::Ptr {
-        archetype.base_pointer::<C>(ty)
+        NonNull::new_unchecked(archetype.base_pointer::<C>(ty))
     }
 
-    fn null() -> Self::Ptr {
-        null_mut()
+    fn dangling() -> Self::Ptr {
+        NonNull::dangling()
     }
 
     unsafe fn deref(ptr: Self::Ptr, idx: u32) -> Self::Item {
-        &mut *ptr.add(idx as usize)
+        &mut *ptr.as_ptr().add(idx as usize)
     }
 }
 
@@ -531,7 +524,7 @@ where
         ty.map(|ty| F::base_pointer(archetype, ty))
     }
 
-    fn null() -> Self::Ptr {
+    fn dangling() -> Self::Ptr {
         None
     }
 
@@ -592,8 +585,8 @@ where
         F::base_pointer(archetype, ty)
     }
 
-    fn null() -> Self::Ptr {
-        F::null()
+    fn dangling() -> Self::Ptr {
+        F::dangling()
     }
 
     unsafe fn deref(ptr: Self::Ptr, idx: u32) -> Self::Item {
@@ -652,8 +645,8 @@ where
         F::base_pointer(archetype, ty)
     }
 
-    fn null() -> Self::Ptr {
-        F::null()
+    fn dangling() -> Self::Ptr {
+        F::dangling()
     }
 
     unsafe fn deref(ptr: Self::Ptr, idx: u32) -> Self::Item {
@@ -708,8 +701,8 @@ macro_rules! impl_fetch_for_tuples {
                 ($($types::base_pointer(archetype, $types),)+)
             }
 
-            fn null() -> Self::Ptr {
-                ($($types::null(),)+)
+            fn dangling() -> Self::Ptr {
+                ($($types::dangling(),)+)
             }
 
             #[allow(non_snake_case)]
@@ -728,7 +721,7 @@ impl_fetch_for_tuples!(A, B, C, D, E, F, G, H, I, J);
 mod tests {
     use super::*;
 
-    use std::mem::forget;
+    use std::mem::{forget, size_of};
 
     fn spawn_three(world: &mut World) {
         let ent = world.alloc();
@@ -929,4 +922,12 @@ mod tests {
             query.get(ent).unwrap();
         }
     }
+
+    #[test]
+    fn fetch_ptr_has_niche() {
+        assert_eq!(
+            size_of::<<<(&i32, &mut f64) as QuerySpec>::Fetch as Fetch>::Ptr>(),
+            size_of::<Option<<<(&i32, &mut f64) as QuerySpec>::Fetch as Fetch>::Ptr>>()
+        );
+    }
 }
